# Copyright 2024 Mattia Giambirtone & All Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import frontend/parsing/lexer


import std/tables

export tables


var tokens* = {"{": TokenType.LeftBrace,
             "}": TokenType.RightBrace,
             "(": TokenType.LeftParen,
             ")": TokenType.RightParen,
             "[": TokenType.LeftBracket,
             "]": TokenType.RightBracket,
             ".": TokenType.Dot,
             ",": TokenType.Comma,
             ":": TokenType.Semicolon,
             "type": TokenType.Type,
             "enum": TokenType.Enum,
             "case": TokenType.Case,
             "operator": TokenType.Operator,
             "generator": TokenType.Generator,
             "fn": TokenType.Function,
             "coroutine": TokenType.Coroutine,
             "break": TokenType.Break,
             "continue": TokenType.Continue,
             "while": TokenType.While,
             "for": TokenType.For,
             "foreach": TokenType.Foreach,
             "if": TokenType.If,
             "else": TokenType.Else,
             "await": TokenType.Await,
             "assert": TokenType.Assert,
             "const": TokenType.Const,
             "let": TokenType.Let,
             "var": TokenType.Var,
             "import": TokenType.Import,
             "yield": TokenType.Yield,
             "return": TokenType.Return,
             "object": TokenType.Object,
             "export": TokenType.Export,
             "block": TokenType.Block,
             "switch": TokenType.Switch,
             "lent": TokenType.Lent,
             "true": TokenType.True,
             "false": TokenType.False,
             "inf": TokenType.Inf,
             "ptr": TokenType.Ptr,
             "nan": TokenType.Nan,
             "inf": TokenType.Inf,
             }.toTable()


for sym in [">", "<", "=", "~", "/", "+", "-", "_", "*", "?", "@", ":", "==", "!=",
            ">=", "<=", "+=", "-=", "/=", "*=", "**=", "!", "%", "&", "|", "^",
            ">>", "<<"]:
                tokens[sym] = TokenType.Symbol


proc fillSymbolTable*(tokenizer: Lexer) =
    ## Initializes the Lexer's symbol
    ## table with builtin symbols and
    ## keywords
    
    # Specialized symbols for which we need a specific token type
    # for easier handling in the parser (it's nicer to use enum members
    # rather than strings whenever possible)
    tokenizer.symbols.addSymbol("{", TokenType.LeftBrace)
    tokenizer.symbols.addSymbol("}", TokenType.RightBrace)
    tokenizer.symbols.addSymbol("(", TokenType.LeftParen)
    tokenizer.symbols.addSymbol(")", TokenType.RightParen)
    tokenizer.symbols.addSymbol("[", TokenType.LeftBracket)
    tokenizer.symbols.addSymbol("]", TokenType.RightBracket)
    tokenizer.symbols.addSymbol(".", TokenType.Dot)
    tokenizer.symbols.addSymbol(",", TokenType.Comma)
    tokenizer.symbols.addSymbol(";", TokenType.Semicolon)

    # Generic symbols avoid us the need to create a gazillion members of the
    # TokenType enum. These are also not handled directly in the parser, but
    # rather processed as classes of operators based on precedence, so using 
    # strings is less of a concern
    for sym in [">", "<", "=", "~", "/", "+", "-", "_", "*", "?", "@", ":", "==", "!=",
            ">=", "<=", "+=", "-=", "/=", "*=", "**=", "!", "%", "&", "|", "^",
            ">>", "<<"]:
        tokenizer.symbols.addSymbol(sym, TokenType.Symbol)
    # Keywords. We differentiate keywords from symbols because they have priority
    # over the latter, and also because the lexer internally uses the symbol map to do
    # maximal matching and it's helpful not to increase the amount of substrings we
    # need to check (especially because keywords match exactly and uniquely, while symbols
    # can share substrings)
    tokenizer.symbols.addKeyword("type", TokenType.Type)
    tokenizer.symbols.addKeyword("enum", TokenType.Enum)
    tokenizer.symbols.addKeyword("case", TokenType.Case)
    tokenizer.symbols.addKeyword("operator", TokenType.Operator)
    tokenizer.symbols.addKeyword("generator", TokenType.Generator)
    tokenizer.symbols.addKeyword("fn", TokenType.Function)
    tokenizer.symbols.addKeyword("coroutine", TokenType.Coroutine)
    tokenizer.symbols.addKeyword("break", TokenType.Break)
    tokenizer.symbols.addKeyword("continue", TokenType.Continue)
    tokenizer.symbols.addKeyword("while", TokenType.While)
    tokenizer.symbols.addKeyword("for", TokenType.For)
    tokenizer.symbols.addKeyword("foreach", TokenType.Foreach)
    tokenizer.symbols.addKeyword("if", TokenType.If)
    tokenizer.symbols.addKeyword("else", TokenType.Else)
    tokenizer.symbols.addKeyword("await", TokenType.Await)
    tokenizer.symbols.addKeyword("assert", TokenType.Assert)
    tokenizer.symbols.addKeyword("const", TokenType.Const)
    tokenizer.symbols.addKeyword("let", TokenType.Let)
    tokenizer.symbols.addKeyword("var", TokenType.Var)
    tokenizer.symbols.addKeyword("import", TokenType.Import)
    tokenizer.symbols.addKeyword("yield", TokenType.Yield)
    tokenizer.symbols.addKeyword("return", TokenType.Return)
    tokenizer.symbols.addKeyword("object", TokenType.Object)
    tokenizer.symbols.addKeyword("export", TokenType.Export)
    tokenizer.symbols.addKeyword("block", TokenType.Block)
    tokenizer.symbols.addKeyword("switch", TokenType.Switch)
    tokenizer.symbols.addKeyword("lent", TokenType.Lent)
    # These are more like expressions with a reserved
    # name that produce a value of a builtin type,
    # but we don't need to care about that until
    # we're in the parsing/compilation steps so
    # it's fine
    tokenizer.symbols.addKeyword("true", TokenType.True)
    tokenizer.symbols.addKeyword("false", TokenType.False)
    tokenizer.symbols.addKeyword("ref", TokenType.Ref)
    tokenizer.symbols.addKeyword("ptr", TokenType.Ptr)
    tokenizer.symbols.addKeyword("nan", TokenType.Nan)
    tokenizer.symbols.addKeyword("inf", TokenType.Inf)
